from limeutils import oxford_comma


class ValidationError(Exception):
    def __init__(self, message='', choices: list = None):
        if choices:
            choices = choices and oxford_comma(choices) or ''
            self.message = message or f'Arguments can only be: {choices}.'
        else:
            self.message = message or 'Arguments must use the correct value or selection of values.'
        super().__init__(self.message)
        
    def __str__(self):
        return self.message


class RedisError(Exception):
    pass





import pytest
from redis.exceptions import ResponseError
from limeutils import byte_conv, ValidationError       # noqa
from icecream import ic     # noqa



param = ['abc', 123, 12.5, 0, 'foo', 'foo', '']
@pytest.mark.parametrize('k', param)
# @pytest.mark.focus
def test_set(red, k):
    assert red.set('sam', k)
    assert red.get('sam') == k
    red.delete('sam')
    

# @pytest.mark.focus
def test_list(red):
    red.delete('many')
    red.set('many', ['a'])
    red.set('many', ['b'])
    red.set('many', ['c'], insert='start')
    red.set('many', ['d'], insert='start')
    assert red.llen(red.formatkey('many')) == 4
    assert (red.get('many')) == ['d', 'c', 'a', 'b']
    red.set('many', ['foo', 'bar'])
    assert red.llen(red.formatkey('many')) == 6
    assert (red.get('many')) == ['d', 'c', 'a', 'b', 'foo', 'bar']
    red.set('many', ['', 'meh'])
    assert red.llen(red.formatkey('many')) == 8
    assert (red.get('many')) == ['d', 'c', 'a', 'b', 'foo', 'bar', '', 'meh']
    red.delete('many')


# @pytest.mark.focus
def test_hash(red):
    red.delete('user')
    red.set('user', dict(age=34, username='enchance', gender='m'))
    assert red.get('user') == dict(age=34, username='enchance', gender='m')
    assert red.get('user', only='username') == dict(username='enchance')
    assert red.get('user', only=['age', 'gender']) == dict(age=34, gender='m')
    red.delete('user')


# @pytest.mark.focus
def test_get_none(red):
    red.delete('xxxyyyzzz')
    assert red.get('xxxyyyzzz') is None
    assert red.get('xxxyyyzzz', 'foo') == 'foo'
    assert red.get('xxxyyyzzz', []) == []
    assert red.get('xxxyyyzzz', [123, 345]) == [123, 345]
    assert red.get('xxxyyyzzz', [123, 345]) == [123, 345]
    assert red.get('xxxyyyzzz', {}) == {}
    assert red.get('xxxyyyzzz', dict(a='b')) == dict(a='b')
    assert red.get('xxxyyyzzz', 123) == 123
    assert red.get('xxxyyyzzz', True)
    assert not red.get('xxxyyyzzz', False)
    


# @pytest.mark.focus
def test_set_data(red):
    red.delete('norepeat')
    red.set('norepeat', {'b', 'a', 'c', 'd', 'a'})
    assert red.get('norepeat') == {'d', 'a', 'b', 'c'}   # unordered of course
    red.delete('norepeat')


param = [
    ('one', 432.5, ['one'], 1), ('two', ['b'], ['one', 'two'], 2),
    ('three', dict(age=34, username='sally', gender='f'), ['one', 'two', 'three'], 3)
]
@pytest.mark.parametrize('key, val, check, out', param)
# @pytest.mark.focus
def test_exists(red, key, val, check, out):
    red.set(key, val)
    assert red.exists(*check) == out
    assert isinstance(out, int)
    
    if check == ['one', 'two', 'three']:
        red.delete(*check)


# @pytest.mark.focus
def test_overwrite(nooverwrite):
    assert nooverwrite.set('a', 'a')
    
    with pytest.raises(ResponseError):
        nooverwrite.set('a', [1])
        assert nooverwrite.set('a', 1)
        assert nooverwrite.set('a', 12.5)
        assert nooverwrite.set('a', [12, 56])
        assert nooverwrite.set('a', dict(foo='bar'))
        
    nooverwrite.delete('a')
    

# @pytest.mark.focus
def test_exception(red):
    with pytest.raises(ValidationError):
        red.set('fail', ['a', 'b'], insert='foo')
        
    try:
        raise ValidationError(choices=['a', 'b'])
    except ValidationError as e:
        assert e.message == 'Arguments can only be: a or b.'
        # ic(e.message)

    try:
        raise ValidationError(choices=['a', 'b', 'c'])
    except ValidationError as e:
        assert e.message == 'Arguments can only be: a, b, or c.'
        # ic(e.message)

    try:
        raise ValidationError(choices=['a'])
    except ValidationError as e:
        assert e.message == 'Arguments can only be: a.'
        # ic(e.message)

    try:
        raise ValidationError('This is it.')
    except ValidationError as e:
        assert e.message == 'This is it.'
        # ic(e.message)
    red.delete('fail')


param = [
    ('ab', 'foo'), ('ab', 1), ('ab', [4, 5]), ('ab', dict(a=1, b=2)),
    (['hello', 'world'], ['foo', 'bar']), (['hello', 'world', 'x'], ['foo', 'bar', 123]),
]
@pytest.mark.parametrize('key, val', param)
# @pytest.mark.focus
# @pytest.mark.skip
def test_delete(red, key, val):
    if isinstance(key, str):
        red.set(key, val)
        v = red.get(key)
        assert v == val
        ret = red.delete(key)
        assert ret == 1
        v = red.get(key)
        assert v is None
        ret = red.delete(key)
        assert ret == 0
    else:
        for idx, k in enumerate(key):
            red.set(k, val[idx])
            assert red.get(k) == val[idx]
        ret = red.delete(*key)
        assert ret == len(key)
        for idx, k in enumerate(key):
            assert red.get(k) is None



from redis import Redis
from typing import Optional, Union, Any, Literal
from pydantic import BaseModel
from redis.client import list_or_args
from redis.exceptions import ResponseError

from . import models
from icecream import ic
from .models import LIST, VAL, StarterModel
from limeutils import byte_conv, ValidationError, listify




class Red(Redis):
    
    def __init__(self, *args, **kwargs):
        self.pre = kwargs.pop('pre', '')
        self.ver = kwargs.pop('ver', '')
        self.ttl = kwargs.pop('ttl', -1)
        self.clear_wrongtype = kwargs.pop('wrongtype', True)
        super().__init__(*args, **kwargs)

    
    def formatkey(self, key: str) -> str:
        """
        Create the final key name with prefix and/or version if necessary
        :param key: The key to format
        :return:    str
        """
        pre = self.pre.strip()
        ver = self.ver.strip()
        
        # Prevent multiple formatting which would be bad
        ll = key.split(':')
        if pre and ver:
            if ll[0] == pre and ll[1] == ver:
                return key
        elif pre and not ver:
            if ll[0] == pre:
                return key
        elif not pre and ver:
            if ll[0] == ver:
                return key

        ll = [pre, ver, key]
        ll = list(filter(None, ll))
        return ":".join(ll)


    def set(self, key: str, val: Union[VAL, LIST, set, dict], **kwargs):
        """
        Set and updates a key
        :param key:     Key name
        :param val:     Value to save. Could be any valid value inc dict or list
        :param kwargs:  Checks for clear, insert, and parent kwargs
        :return:        Depends on the type of key
        """
        key = self.formatkey(key)
        clear = kwargs.pop('clear', False)
        insert = kwargs.pop('insert', 'end')
        ex = kwargs.pop('ttl', self.ttl)
        
        if clear:
            self.delete(key)
        
        if isinstance(val, (str, int, float, bytes)):
            if self.clear_wrongtype and self._get_type(key) != 'string':
                self.delete(key)
            return super().set(key, val, ex=ex, **kwargs)
        
        elif isinstance(val, (list, tuple)):
            if self.clear_wrongtype and self._get_type(key) != 'list':
                self.delete(key)
                    
            if insert == 'end':
                ret = self.rpush(key, *val)
            elif insert == 'start':
                ret = self.lpush(key, *val)
            else:
                raise ValidationError(choices=['start', 'end'])
            
            if ex != -1:
                self.expire(key, ex)
            return ret
            
        elif isinstance(val, dict):
            if self.clear_wrongtype and self._get_type(key) != 'hash':
                self.delete(key)
            ret = self.hset(key, mapping=val)
            if ex != -1:
                self.expire(key, ex)
            return ret
        
        elif isinstance(val, set):
            if self.clear_wrongtype and self._get_type(key) != 'set':
                self.delete(key)
            ret = self.sadd(key, *val)
            if ex != -1:
                self.expire(key, ex)
            return ret
    
    
    def get(self, key: str, default=None, **kwargs):
        """
        Get the value of a key
        :param key:     Key name
        :param default: Default value if key doesn't exist
        :param kwargs:  Checks for start, end, only, and parent kwargs
        :return:        Depends on the type of key
        """
        start = kwargs.pop('start', 0)
        end = kwargs.pop('end', -1)
        only = kwargs.pop('only', None)
        
        key = self.formatkey(key)
        datatype = byte_conv(super().type(key))
        
        if datatype == 'string':
            return byte_conv(super().get(key))
        
        elif datatype == 'list':
            data = super().lrange(key, start, end)
            return [byte_conv(i) for i in data]
        
        elif datatype == 'hash':
            if only:
                only = [only] if isinstance(only, str) else list(only)
                data = super().hmget(key, only)
                data = [byte_conv(i) for i in data]
                d = dict(zip(only, data))
            else:
                data = super().hgetall(key)
                d = {byte_conv(k):byte_conv(v) for k, v in data.items()}
            # ic(d)
            return d
        elif datatype == 'set':
            data = super().smembers(key)
            return {byte_conv(v) for v in data}
        return default


    def exists(self, *keys) -> int:
        """
        Checks if keys exists. Returns the number of keys that exist not which of the keys exists.
        :param keys:    Names of keys
        :return:        int No. of keys that exist
        """
        for i in keys:
            pass
        keys = [self.formatkey(i) for i in keys]
        return super().exists(*keys)
    
    
    def delete(self, *keys):
        """
        Delete keys
        :param keys:    Keys to delete
        :return:        int Number of keys deleted
        """
        keys = list(map(self.formatkey, keys))
        return super().delete(*keys)

    
    def _get_type(self, key: str):
        return byte_conv(super().type(key))


# class Redis:
#     def __init__(self, **kwargs):
#         self.conn = reds.Redis(**kwargs)
#
#
#
#
#
#     def hset(self, key: str, field: str, val: Optional[V] = '', mapping: Optional[dict] = None,
#              ttl=None, pre=None, ver=None) -> int:
#         data = models.Hset(key=key, field=field, val=val, mapping=mapping,
#                            ttl=ttl, pre=pre, ver=ver)
#         key = self.cleankey(data)
#         ttl = data.ttl if data.ttl is not None else self.ttl
#
#         self.pipe.hset(key, data.field, data.val, data.mapping)
#         self.pipe.expire(key, ttl)
#         [hset_ret, _] = self.pipe.execute()
#         return hset_ret


from typing import Optional, Union, Any
from pydantic import BaseModel, Field, validator


VAL = Union[str, int, float, bytes]
LIST = Union[list, tuple]

#
#
# def listmaker(val):
#     """Validator: Convert str to list with one item."""
#     if isinstance(val, str):
#         return [val]
#     return val
#
# def nonone(val):
#     """Validator: Convert None to empty string."""
#     if val is None:
#         return ''
#     return val
#
# def nonone_mapping(val):
#     """Validator: Convert None to empty string."""
#     if val is None or not val:
#         return None
#     else:
#         for k, v in val.items():
#             val[k] = '' if v is None else v
#     return val


class StarterModel(BaseModel):
    """Shared fields"""
    key: VAL
    pre: Optional[VAL] = ''
    ver: Optional[VAL] = ''
    ttl: Optional[int] = Field(-1, ge=0)


# class Hset(StarterModel):
#     """Pydantic model for the hset() method"""
#     field: str
#     val: Optional[VAL]
#     mapping: Optional[dict] = None
#
#     _clean_val = validator('val', allow_reuse=True)(nonone)
#     _clean_mapping = validator('mapping', allow_reuse=True)(nonone_mapping)
#
#
# class Hmset(StarterModel):
#     """Pydantic model for the hmset() method"""
#     mapping: Optional[dict] = None
#
#     _clean_mapping = validator('mapping', allow_reuse=True)(nonone_mapping)
#
#
# class Set(StarterModel):
#     """Pydantic model for the set() method"""
#     val: Optional[VAL] = ''
#     xx: bool = False
#     keepttl: bool = False
#
#     _clean_val = validator('val', allow_reuse=True)(nonone)
#
#
#     @validator('xx', 'keepttl')
#     def boolonly(cls, val):
#         return bool(val)
#
#
# class Get(StarterModel):
#     """Pydantic model for the get() method"""
#     default: Optional[Any] = ''
#
#
# class Hget(StarterModel):
#     """Pydantic model for the get() method"""
#     default: Optional[Any] = ''
#
#
# class Hmget(StarterModel):
#     """Pydantic model for the hmget() method"""
#     fields_: Optional[LIST] = None
#
#
# class Hdel(StarterModel):
#     """Pydantic model for the hdel() method"""
#     fields_: Optional[Union[str, LIST]] = None
#
#     _clean_fields = validator('fields_', allow_reuse=True)(listmaker)
#
#
# class Delete(BaseModel):
#     """Pydantic model for the delete() method"""
#     key: Union[str, LIST]
#     pre: Optional[VAL] = ''
#     ver: Optional[VAL] = ''
#
#     _clean_fields = validator('key', allow_reuse=True)(listmaker)




import importlib


def model_str(instance, attr) -> str:
    """
    The field to display for an object's __str__. If the field doesn't exist then an
    alternative will be displayed.
    :param instance:    Instance object
    :param attr:        Field name to get data from if it exists
    :return:            str
    """
    return hasattr(instance, attr) and getattr(instance, attr) \
           or f'<{instance.__class__.__name__}: {instance.id}>'


def modstr(instance, attr) -> str:
    """Alias for model_str()"""
    return model_str(instance, attr)


def classgrabber(dotpath: str):
    """
    Returns the class from a dot path that leads to the class to be imported.
    The class would be ready for use.
    Example:
        Settings = classgrabber('app.folder.file.Settings')
        # Settings class now ready for use
        myobj = Settings()
    :param dotpath:   A dot path
    :return:            class
    """
    x = dotpath.split('.')
    path = '.'.join(x[0:-1])
    models = importlib.import_module(path)
    return getattr(models, x[-1])


import pytest
from limeutils import Red





@pytest.fixture(scope='session')
def red():
    CACHE_CONFIG = {
        # 'host': 'localhost',
        # 'port': 6379,
        'db': 1,
        'pre': 'TEST',
        'ver': '',
        'ttl': 3600 * 24 * 15,
    }
    yield Red(**CACHE_CONFIG)


@pytest.fixture(scope='session')
def nooverwrite():
    CACHE_CONFIG = {
        # 'host': 'localhost',
        # 'port': 6379,
        'db': 1,
        'pre': 'TEST',
        'ver': '',
        'ttl': 3600 * 24 * 15,
        'wrongtype': False
    }
    return Red(**CACHE_CONFIG)








#
#
# @pytest.fixture(scope='session', autouse=True)
# def delete_keys(r):
#     # print('FOO')
#     yield
#     r.delete(['football', 'foobar', 'mustard', 'abra', 'sam', 'rrr', 'lll', 'qqq', 'www'])
#     # print('BAR')
#
# def data(r):
#     # hmget
#     r.hset('abra', 'foo', 'bar')
#     r.hset('abra', 'fed', 23)
#     r.hset('abra', 'meh', 5.2)
#     r.hset('abra', 'nothing', 0)
#     r.hset('abra', 'zoom', None)
#
#     # hdel
#     r.hmset('mustard', dict(aaa=1, bbb=2, ccc=3, ddd=4, eee=5))
#
#     # delete
#     r.hset('aaa1', 'abc', 23)
#     r.hset('aaa2', 'abc', 23)
#     r.hset('aaa3', 'abc', 23)
#     r.hset('aaa4', 'abc', 23)
#     r.hset('aaa5', 'abc', 23)
#
#     # get
#     r.set('qqq', 'abc')
#     r.set('www', '123')
#     r.set('rrr', '')
#     r.set('lll', None)


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="limeutils", # Replace with your own username
    version="0.2.4",
    author="dropickdev",
    author_email="enchance@gmail.com",
    description="Utility functions for python. Functions were made and tested in FastAPI but are "
                "suitable for any python project.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dropkickdev/limeutils.git",
    packages=setuptools.find_packages(),
    install_requires=['redis', 'pydantic'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


from typing import Optional, Union
from ast import literal_eval


def isfloat(val: Union[int, float, str]):
    """
    Checks if the contents of a string is a float.
    :param val: String to parse
    :return:    bool
    """
    try:
        if isinstance(val, str):
            val.strip()
        float(val)
        return True
    except ValueError:
        return False
    

def byte_conv(val: Union[bytes, str]):
    """
    Converts bytes to a python string. This string could later be parsed into the correct
    python data type using parse_str(). Used mostly with Redis return values which always return in
    bytes.
    :param val: Bytes string to convert to python string
    :return:    str
    """
    if isinstance(val, str):
        return val
    
    try:
        x = val.decode()
        x = isinstance(x, bytes) and len(x) == 0 and '' or parse_str(x)
        return x
    except (UnicodeDecodeError, AttributeError):
        return val
    # return val.decode('utf-8')


def parse_str(string: str):
    """
    Converts a string to either an int, float, or str depending on its value.
    :param string:   String to convert
    :return:    int, float, or str
    """
    if not isinstance(string, str):
        raise ValueError('Only valid strings can be parsed.')
    
    string = string.strip()
    if string.isdigit():
        return int(string)
    elif isfloat(string):
        return float(string)
    elif string in ['True', 'False']:
        return literal_eval(string)
    return string


def split_fullname(fullname: str, default: str = '',
                   prefix: Optional[Union[str, list, tuple]] = None,
                   suffix: Optional[Union[str, list, tuple]] = None) -> tuple:
    """
    Splits a fullname into their respective first_name and last_name fields.
    If only one name is given, that becomes the first_name
    :param fullname:    The name to split
    :param default:     The value if only one name is given
    :param prefix:      Custom prefixes to append to the default list
    :param suffix:      Custom suffixes to append to the default list
    :return:            tuple
    """
    if prefix and not isinstance(prefix, (str, list, tuple)):
        raise TypeError('`prefix` must be a list/str for multi/single values.')

    if suffix and not isinstance(suffix, (str, list, tuple)):
        raise TypeError('`suffix` must be a list/str for multi/single values.')

    prefix = isinstance(prefix, str) and [prefix] or prefix or []
    suffix = isinstance(suffix, str) and [suffix] or suffix or []
    PREFIX_LASTNAME = ['dos', 'de', 'delos', 'san', 'dela', 'dona', *prefix]
    SUFFIX_LASTNAME = ['phd', 'md', 'rn', *suffix]

    list_ = fullname.split()
    lastname_idx = None
    if len(list_) > 2:
        for idx, val in enumerate(list_):
            if val.lower() in PREFIX_LASTNAME:
                lastname_idx = idx
                break
            elif val.lower().replace('.', '') in SUFFIX_LASTNAME:
                lastname_idx = idx - 1
            else:
                if idx == len(list_) - 1:
                    lastname_idx = idx
                else:
                    continue
        list_[:lastname_idx] = [' '.join(list_[:lastname_idx])]
        list_[1:] = [' '.join(list_[1:])]
    try:
        first, last = list_
    except ValueError:
        first, last = [*list_, default]
    return first, last


def oxford_comma(sequence: Union[list, tuple, set], separator: str = 'or'):
    sequence = sequence and list(sequence) or []
    if not sequence:
        return ''
    elif len(sequence) == 1:
        return sequence[0]
    elif len(sequence) == 2:
        return '{} {} {}'.format(', '.join(sequence[:-1]), separator, sequence[-1])
    elif len(sequence) >= 3:
        return '{}, {} {}'.format(', '.join(sequence[:-1]), separator, sequence[-1])
    
    
def listify(data: Union[str, int, float, list, set, tuple, bool]) -> list:
    """
    A convenience function that converts data into a list unless it's already one.
    Bool is a subset of int.
    :param data:    Data to turn into a list
    :return:        list
    """
    if isinstance(data, list):
        return data
    elif isinstance(data, (set, tuple)):
        return list(data)
    else:
        return [data]


# TESTME: Untested
def valid_str_only(item, allow_bool: bool = False) -> bool:
    """
    Removes empty strings and other invalid values passed that aren't valid strings.
    :param item:        Check if valid str
    :param allow_bool:  Allow the use of bool only if True
    :return:            bool
    """
    if isinstance(item, str) and len(item) > 0:
        return True
    if allow_bool and item:
        return True
    return False


import pytest
from collections import Counter
from icecream import ic
from limeutils import utils, listify, valid_str_only


param = [(3, True), (3.0, True), (0, True), ('3.4', True), ('0.4', True), ('0.0', True),
         ('0', True), ('3.0', True), ('3', True), ('abc', False), ('3,344', False),
         ('3,344.5', False), ('3,344.00', False)]
@pytest.mark.parametrize('val, out', param)
def test_isfloat(val, out):
    assert utils.isfloat(val) is out


param = [('123', 123), ('12.3', 12.3), ('a1b2c3', 'a1b2c3'), ('abc', 'abc'), ('', ''), ('-', '-'),
         (None, ValueError), (bytes(), ValueError), ('True', True), ('False', False),]
@pytest.mark.parametrize('val, out', param)
def test_parse_str(val, out):
    try:
        assert utils.parse_str(val) == out
    except out:
        with pytest.raises(out):
            assert utils.parse_str(val)


param = [('Hey You', ('Hey', 'You')), ('Sir Hey You', ('Sir Hey', 'You')),
         ('Sir Hey You Phd', ('Sir Hey', 'You Phd')), ('Hey delos You', ('Hey', 'delos You')),
         ('Hey san You', ('Hey', 'san You')),
         ('Eliza Maria Erica dona Aurora Phd Md', ('Eliza Maria Erica', 'dona Aurora Phd Md'))]
@pytest.mark.parametrize('val, out', param)
def test_split_fullname(val, out):
    assert utils.split_fullname(val) == out


param = [('abra', 'fed'), ('abra', 6.22), ('abra', 789), ('abra', '1.5'), ('abra', '123')]
@pytest.mark.parametrize('key, val', param)
def test_byte_conv(red, key, val):
    red.set(key, val)
    if val == '123':
        val = 123
    if val == '1.5':
        val = 1.5
    assert red.get(key) == val


param = [
    (['one', 'two', 'three', 'four'], 'one, two, three, or four'),
    (['one', 'two', 'three'], 'one, two, or three'),
    (['one', 'two'], 'one or two'),
    (['one'], 'one'), ([], '')
]
@pytest.mark.parametrize('seq, out', param)
@pytest.mark.utilfocus
def test_oxford_comma(seq, out):
    assert utils.oxford_comma(seq) == out


param = [
    ('foo', ['foo']), (['foo'], ['foo']),
    (1, [1]), (12.5, [12.5]),
    (['foo', 'bar'], ['foo', 'bar']),
    (('foo',), ['foo']), (('foo', 'bar'), ['foo', 'bar']),
    ({'foo'}, ['foo']), ({'foo', 'bar'}, ['foo', 'bar']),
    (True, [True]), (False, [False])
]
@pytest.mark.parametrize('data, out', param)
# @pytest.mark.focus
def test_listify(data, out):
    assert Counter(listify(data)) == Counter(out)


param = [
    ('', False, False), ([], False, False), (None, False, False), (1, False, False),
    (1.5, False, False), (False, False, False), (True, False, False), ('a', False, True),
    (set(), False, False), ('', False, False), (True, True, True), (False, True, False)
]
@pytest.mark.parametrize('item, allow, out', param)
@pytest.mark.focus
def test_valid_str_only(item, allow, out):
    if allow:
        assert valid_str_only(item, allow_bool=True) == out
    else:
        assert valid_str_only(item) == out
        